@@ -28,7 +28,10 @@ import subprocess
 import shutil
 import calendar
 
-from fcntl import flock, LOCK_EX, LOCK_UN, LOCK_NB
+import threading
+import Queue
+
+from fcntl import lockf, LOCK_EX, LOCK_UN, LOCK_NB
 from StringIO import StringIO
 
 from git import Repo
@@ -36,6 +39,35 @@ from git import Repo
 # default basic logger. We override it later.
 logger = logging.getLogger(__name__)
 
+class PullerThread(threading.Thread):
+    def __init__(self, in_queue, out_queue, config, thread_name):
+        threading.Thread.__init__(self)
+        self.in_queue   = in_queue
+        self.out_queue  = out_queue
+        self.toplevel   = config['toplevel']
+        self.hookscript = config['post_update_hook']
+        self.myname     = thread_name
+
+    def run(self):
+        while True:
+            (gitdir, modified) = self.in_queue.get()
+            fullpath = os.path.join(self.toplevel, gitdir.lstrip('/'))
+            try:
+                grokmirror.lock_repo(fullpath, nonblocking=True)
+                logger.info('[Thread-%s] Updating %s' % (self.myname, gitdir))
+                pull_repo(self.toplevel, gitdir)
+                set_agefile(self.toplevel, gitdir, modified)
+                run_post_update_hook(self.hookscript, self.toplevel, gitdir)
+                grokmirror.unlock_repo(fullpath)
+                self.out_queue.put((gitdir, True))
+            except IOError, ex:
+                logger.info('Could not obtain exclusive lock on %s' % gitdir)
+                logger.info('\tAssuming grok-fsck is running, will try later.')
+                self.out_queue.put((gitdir, False))
+            logger.debug('[Thread-%s] done pulling %s' % (self.myname, gitdir))
+
+            self.in_queue.task_done()
+
 def fix_remotes(gitdir, toplevel, site):
     # Remove all existing remotes and set new origin
     repo = Repo(os.path.join(toplevel, gitdir.lstrip('/')))
@@ -70,7 +102,7 @@ def set_owner_description(toplevel, gitdir, owner, description):
 def set_agefile(toplevel, gitdir, last_modified):
     # set agefile, which can be used by cgit to show idle times
     # cgit recommends it to be yyyy-mm-dd hh:mm:ss
-    cgit_fmt = time.strftime('%F %T', time.gmtime(last_modified))
+    cgit_fmt = time.strftime('%F %T', time.localtime(last_modified))
     agefile = os.path.join(toplevel, gitdir.lstrip('/'),
             'info/web/last-modified')
     if not os.path.exists(os.path.dirname(agefile)):
@@ -106,7 +138,6 @@ def run_post_update_hook(hookscript, toplevel, gitdir):
 def pull_repo(toplevel, gitdir):
     env = {'GIT_DIR': os.path.join(toplevel, gitdir.lstrip('/'))}
     args = ['/usr/bin/git', 'remote', 'update', '--prune']
-    logger.info('Updating %s' % gitdir)
 
     logger.debug('Running: GIT_DIR=%s %s' % (env['GIT_DIR'], ' '.join(args)))
 
@@ -213,6 +244,11 @@ def write_projects_list(manifest, config):
     if 'projectslist_trimtop' in config.keys():
         trimtop = config['projectslist_trimtop']
 
+    add_symlinks = False
+    if ('projectslist_symlinks' in config.keys()
+            and config['projectslist_symlinks'] == 'yes'):
+        add_symlinks = True
+
     (dirname, basename) = os.path.split(plpath)
     (fd, tmpfile) = tempfile.mkstemp(prefix=basename, dir=dirname)
     logger.info('Writing new %s' % plpath)
@@ -221,11 +257,23 @@ def write_projects_list(manifest, config):
         fh = open(tmpfile, 'w')
         for gitdir in manifest.keys():
             if trimtop and gitdir.find(trimtop) == 0:
-                gitdir = gitdir[len(trimtop):]
+                pgitdir = gitdir[len(trimtop):]
+            else:
+                pgitdir = gitdir
 
             # Always remove leading slash, otherwise cgit breaks
-            gitdir = gitdir.lstrip('/')
-            fh.write('%s\n' % gitdir)
+            pgitdir = pgitdir.lstrip('/')
+            fh.write('%s\n' % pgitdir)
+
+            if add_symlinks and 'symlinks' in manifest[gitdir].keys():
+                # Do the same for symlinks
+                # XXX: Should make this configurable, perhaps
+                for symlink in manifest[gitdir]['symlinks']:
+                    if trimtop and symlink.find(trimtop) == 0:
+                        symlink = symlink[len(trimtop):]
+
+                    symlink = symlink.lstrip('/')
+                    fh.write('%s\n' % symlink)
 
         fh.close()
         os.chmod(tmpfile, 0644)
@@ -236,7 +284,6 @@ def write_projects_list(manifest, config):
         if os.path.exists(tmpfile):
             os.unlink(tmpfile)
 
-
 def pull_mirror(name, config, opts):
     global logger
     logger = logging.getLogger(name)
@@ -244,7 +291,7 @@ def pull_mirror(name, config, opts):
 
     if 'log' in config.keys():
         ch = logging.FileHandler(config['log'])
-        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
+        formatter = logging.Formatter("[%(process)d] %(asctime)s - %(levelname)s - %(message)s")
         ch.setFormatter(formatter)
         loglevel = logging.INFO
 
@@ -269,72 +316,99 @@ def pull_mirror(name, config, opts):
     # push it into grokmirror to override the default logger
     grokmirror.logger = logger
 
-    logger.info('Updating mirror for [%s]' % name)
-
     # Lock the tree to make sure we only run one instance
     logger.debug('Attempting to obtain lock on %s' % config['lock'])
     flockh = open(config['lock'], 'w')
     try:
-        flock(flockh, LOCK_EX | LOCK_NB)
+        lockf(flockh, LOCK_EX | LOCK_NB)
     except IOError, ex:
-        logger.info('Could not obtain exclusive lock on %s' % config['lock'])
-        logger.info('Assuming another process is running.')
+        logger.debug('Could not obtain exclusive lock on %s' % config['lock'])
+        logger.info('Skipping [%s] - update already running' % name)
         return 0
 
+    logger.info('Checking [%s]' % name)
+
     mymanifest = config['mymanifest']
-    logger.info('Fetching remote manifest from %s' % config['manifest'])
-    request = urllib2.Request(config['manifest'])
-    opener  = urllib2.build_opener()
 
-    # Find out if we need to run at all first
-    if not (opts.force or opts.nomtime) and os.path.exists(mymanifest):
-        fstat = os.stat(mymanifest)
-        mtime = fstat[8]
-        logger.debug('mtime on %s is: %s' % (mymanifest, mtime))
-        my_last_modified = time.strftime('%a, %d %b %Y %H:%M:%S GMT',
-            time.gmtime(mtime))
-        logger.debug('Our last-modified is: %s' % my_last_modified)
-        request.add_header('If-Modified-Since', my_last_modified)
+    if config['manifest'].find('file:///') == 0:
+        manifile = config['manifest'].replace('file://', '')
+        if not os.path.exists(manifile):
+            logger.critical('Remote manifest not found in %s! Quitting!' % config['manifest'])
+            return 1
 
-    try:
-        ufh = opener.open(request)
-    except urllib2.HTTPError, ex:
-        if ex.code == 304:
-            logger.info('Server says we have the latest manifest. Quitting.')
-            flock(flockh, LOCK_UN)
-            flockh.close()
-            return 0
-        logger.critical('Could not fetch %s' % config['manifest'])
-        logger.critical('Server returned: %s' % ex)
-        flock(flockh, LOCK_UN)
-        flockh.close()
-        return 1
-    except urllib2.URLError, ex:
-        logger.critical('Could not fetch %s' % config['manifest'])
-        logger.critical('Error was: %s' % ex)
-        flock(flockh, LOCK_UN)
-        flockh.close()
-        return 1
+        fstat = os.stat(manifile)
+        last_modified = fstat[8]
+        if os.path.exists(config['mymanifest']):
+            fstat = os.stat(config['mymanifest'])
+            my_last_modified = fstat[8]
+            if not (opts.force or opts.nomtime) and last_modified <= my_last_modified:
+                logger.info('Manifest file unchanged. Quitting.')
+                lockf(flockh, LOCK_UN)
+                flockh.close()
+                return 0
 
-    last_modified = ufh.headers.get('Last-Modified')
-    last_modified = time.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
-    last_modified = calendar.timegm(last_modified)
+        logger.info('Reading new manifest from %s' % manifile)
+        manifest = grokmirror.read_manifest(manifile)
+        # Don't accept empty manifests -- that indicates something is wrong
+        if not len(manifest.keys()):
+            logger.critical('Remote manifest empty or unparseable! Quitting.')
+            return 1
 
-    # We don't use read_manifest for the remote manifest, as it can be
-    # anything, really. For now, blindly open it with gzipfile if it ends
-    # with .gz. XXX: some http servers will auto-deflate such files.
-    if config['manifest'].find('.gz') > 0:
-        fh = gzip.GzipFile(fileobj=StringIO(ufh.read()))
     else:
-        fh = ufh
+        # Load it from remote host using http and header magic
+        logger.info('Fetching remote manifest from %s' % config['manifest'])
+        request = urllib2.Request(config['manifest'])
+        opener  = urllib2.build_opener()
 
-    try:
-        manifest = json.load(fh)
-    except:
-        logger.critical('Failed to parse %s' % config['manifest'])
-        flock(flockh, LOCK_UN)
-        flockh.close()
-        return 1
+        # Find out if we need to run at all first
+        if not (opts.force or opts.nomtime) and os.path.exists(mymanifest):
+            fstat = os.stat(mymanifest)
+            mtime = fstat[8]
+            logger.debug('mtime on %s is: %s' % (mymanifest, mtime))
+            my_last_modified = time.strftime('%a, %d %b %Y %H:%M:%S GMT',
+                time.gmtime(mtime))
+            logger.debug('Our last-modified is: %s' % my_last_modified)
+            request.add_header('If-Modified-Since', my_last_modified)
+
+        try:
+            ufh = opener.open(request, timeout=30)
+        except urllib2.HTTPError, ex:
+            if ex.code == 304:
+                logger.info('Server says we have the latest manifest. Quitting.')
+                lockf(flockh, LOCK_UN)
+                flockh.close()
+                return 0
+            logger.critical('Could not fetch %s' % config['manifest'])
+            logger.critical('Server returned: %s' % ex)
+            lockf(flockh, LOCK_UN)
+            flockh.close()
+            return 1
+        except urllib2.URLError, ex:
+            logger.critical('Could not fetch %s' % config['manifest'])
+            logger.critical('Error was: %s' % ex)
+            lockf(flockh, LOCK_UN)
+            flockh.close()
+            return 1
+
+        last_modified = ufh.headers.get('Last-Modified')
+        last_modified = time.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
+        last_modified = calendar.timegm(last_modified)
+
+        # We don't use read_manifest for the remote manifest, as it can be
+        # anything, really. For now, blindly open it with gzipfile if it ends
+        # with .gz. XXX: some http servers will auto-deflate such files.
+        if config['manifest'].find('.gz') > 0:
+            fh = gzip.GzipFile(fileobj=StringIO(ufh.read()))
+        else:
+            fh = ufh
+
+        try:
+            manifest = json.load(fh)
+        except:
+            logger.critical('Failed to parse %s' % config['manifest'])
+            lockf(flockh, LOCK_UN)
+            flockh.close()
+            return 1
 
     mymanifest = grokmirror.read_manifest(mymanifest)
 
@@ -365,7 +439,7 @@ def pull_mirror(name, config, opts):
 
     toplevel = config['toplevel']
     if not os.access(toplevel, os.W_OK):
-        logger.critical('Toplevel %s does not exist or is not writable')
+        logger.critical('Toplevel %s does not exist or is not writable' % toplevel)
         sys.exit(1)
 
     for gitdir in culled.keys():
@@ -412,20 +486,41 @@ def pull_mirror(name, config, opts):
 
     hookscript = config['post_update_hook']
 
+    in_queue  = Queue.Queue()
+    out_queue = Queue.Queue()
+
+    # start out the number of threads we want
+    if 'pull_threads' in config.keys():
+        pull_threads = int(config['pull_threads'])
+        if pull_threads < 1:
+            logger.info('The pull_threads option is less than 1, forcing to 1')
+            pull_threads = 1
+    else:
+        # be conservative
+        logger.info('The pull_threads option is not set, consider setting it')
+        pull_threads = 5
+
+    logger.info('Will use %d threads to pull repos' % pull_threads)
+
     for gitdir in to_pull:
-        # Check in case grok-fsck is checking this repo right now
-        try:
-            fullpath = os.path.join(toplevel, gitdir.lstrip('/'))
-            grokmirror.lock_repo(fullpath, nonblocking=True)
-            pull_repo(toplevel, gitdir)
-            set_agefile(toplevel, gitdir, culled[gitdir]['modified'])
-            run_post_update_hook(hookscript, toplevel, gitdir)
-            grokmirror.unlock_repo(fullpath)
-        except IOError, ex:
-            logger.info('Could not obtain exclusive lock on %s' % gitdir)
-            logger.info('\tAssuming grok-fsck is running, will try later.')
+        in_queue.put((gitdir, culled[gitdir]['modified']))
+
+    for i in range(pull_threads):
+        t = PullerThread(in_queue, out_queue, config, i)
+        t.setDaemon(True)
+        t.start()
+
+    # wait till it's all done
+    in_queue.join()
+    logger.info('All threads finished.')
+
+    while not out_queue.empty():
+        # see if any of it failed
+        (gitdir, status) = out_queue.get()
+        if status == False:
             # To make sure we check this again during next run,
             # fudge the manifest accordingly.
+            logger.debug('Will recheck %s during next run' % gitdir)
             culled[gitdir] = mymanifest[gitdir]
             # this is rather hackish, but effective
             last_modified -= 1
@@ -469,7 +564,7 @@ def pull_mirror(name, config, opts):
                 if symlink not in symlinks:
                     symlinks.append(symlink)
                 target = os.path.join(config['toplevel'], symlink.lstrip('/'))
-                if not os.path.exists(target):
+                if not os.path.exists(target) and os.path.exists(source):
                     logger.info('Symlinking %s -> %s' % (target, source))
                     # Make sure the leading dirs are in place
                     if not os.path.exists(os.path.dirname(target)):
@@ -488,13 +583,14 @@ def pull_mirror(name, config, opts):
                     shutil.rmtree(founddir)
 
     # Once we're done, save culled as our new manifest
-    grokmirror.write_manifest(config['mymanifest'], culled, last_modified)
+    grokmirror.write_manifest(config['mymanifest'], culled,
+                              mtime=last_modified, pretty=opts.pretty)
 
     # write out projects.list, if asked to
     write_projects_list(culled, config)
 
     logger.debug('Unlocking %s' % config['lock'])
-    flock(flockh, LOCK_UN)
+    lockf(flockh, LOCK_UN)
     flockh.close()
 
     return 127
@@ -522,6 +618,9 @@ if __name__ == '__main__':
     parser.add_option('-p', '--purge', dest='purge',
         action='store_true', default=False,
         help='Remove any git trees that are no longer in manifest.')
+    parser.add_option('-y', '--pretty', dest='pretty', action='store_true',
+        default=False,
+        help='Pretty-print manifest (sort keys and add indentation)')
     parser.add_option('-r', '--reuse-existing-repos', dest='reuse',
         action='store_true', default=False,
         help='If any existing repositories are found on disk, set new '
@@ -531,6 +630,9 @@ if __name__ == '__main__':
 
     (opts, args) = parser.parse_args()
 
+    if not opts.config:
+        parser.error('You must provide the path to the config file')
+
     ini = ConfigParser()
     ini.read(opts.config)
 
@@ -545,6 +647,10 @@ if __name__ == '__main__':
             config['default_owner'] = 'Grokmirror User'
         if 'post_update_hook' not in config.keys():
             config['post_update_hook'] = ''
+        if 'include' not in config.keys():
+            config['include'] = '*'
+        if 'exclude' not in config.keys():
+            config['exclude'] = ''
 
         sect_retval = pull_mirror(section, config, opts)
         if sect_retval == 1:
